#include "MeshLoader.h"
#include "NeuMeshLoader.h"
#include "AneuMeshLoader.h"
#include "Exceptions.h"
#include <cmath>
#include <ranges>
#include <iostream>
#include <string>
#include <fstream>
#include <algorithm>


//!!! Задание на защиту (не удалять!)
//!!! Используя std::ranges реализовать:
//!!! 1. Получить ID кэ с данной грани (поверхности)
//!!! 2. Среди них найти КЭ с минимальной площадью.

struct sides {
    double ab=0,ac=0,bc=0;
    sides(Node& a, Node& b, Node& c) {
        double x1 = a.x, y1 = a.y, z1 = a.z,
            x2 = b.x, y2 = b.y, z2 = b.z,
            x3 = c.x, y3 = c.y, z3 = c.z;
        ab = sqrt((x1-x2)*(x1 - x2) +(y1-y2)*(y1 - y2) +(z1-z2)*(z1 - z2));
        ac = sqrt((x1 - x3) * (x1 - x3) + (y1 - y3) * (y1 - y3) + (z1 - z3) * (z1 - z3));
        bc = sqrt((x3 - x2) * (x3 - x2) + (y3 - y2) * (y3 - y2) + (z3 - z2) * (z3 - z2));
    };
};

double square(MeshLoader& MyKE,int id) {
    Node a = MyKE.Get_Nodes().at(MyKE.Get_Elements().at(id-1).nodes_id.at(0)-1),
        b = MyKE.Get_Nodes().at(MyKE.Get_Elements().at(id-1).nodes_id.at(1)-1),
        c = MyKE.Get_Nodes().at(MyKE.Get_Elements().at(id-1).nodes_id.at(2)-1),
        d = MyKE.Get_Nodes().at(MyKE.Get_Elements().at(id-1).nodes_id.at(3)-1);
    sides elem1(a, b, c), elem2(a, b, d), elem3(a, c, d), elem4(b, c, d);
    double P1 = (elem1.ab + elem1.ac + elem1.bc) / 2,
        P2 = (elem2.ab + elem2.ac + elem2.bc) / 2,
        P3 = (elem3.ab + elem3.ac + elem3.bc) / 2,
        P4 = (elem4.ab + elem4.ac + elem4.bc) / 2;
        
    return sqrt((P1 - elem1.ab) * (P1 - elem1.ac) * (P1 - elem1.bc) * P1) + sqrt((P2 - elem2.ab) * (P2 - elem2.ac) * (P2 - elem2.bc) * P2) +
        sqrt((P3 - elem3.ab) * (P3 - elem3.ac) * (P3 - elem3.bc) * P3) + sqrt((P4 - elem4.ab) * (P4 - elem4.ac) * (P4 - elem4.bc) * P4);
}

void Protection_task(MeshLoader& MyKE) {
    int id;
    std::cout << "Enter border ID:";
    std::cin >> id;
    std::set<int> Container = MyKE.Get_Elements_by_ID_area(id);


    //!!! Для поиска минимального элемента есть ranges::min_element, принимающий бинарный предикат,
    //!!! в данном случае для сревнения площадей


    static double min=square(MyKE,*Container.begin());
    auto min_square = [&](auto& elem) {if (square(MyKE, elem) < min)min = square(MyKE, elem);
    return square(MyKE, elem) <= min; };
    auto alg= Container | std::ranges::views::filter(min_square);
    int result=0;
    for (auto& i : alg) 
        result = i;
    std::cout << result;
}


MeshLoader* File_Format_Checker(const std::string& _f) 
{ 
	std::string file(_f);
    while (file.back() == ' ')                      
        file.pop_back();
    MeshLoader* mesh_pointer = nullptr;

    std::size_t found = file.rfind(".aneu");
    if (found != std::string::npos && file.length() - found == 5) 
    {
        std::cout << "File format passed validation!" << std::endl;
        
        mesh_pointer = new AneuMeshLoader;
        
        return mesh_pointer;
    }

    found = file.rfind(".neu");
    if (found != std::string::npos && file.length() - found == 4) 
    {
        std::cout << "File format passed validation!" << std::endl;
        mesh_pointer = new NeuMeshLoader;
        return mesh_pointer;
    }
    else 
    {
        std::cout << "WARNING! File format failed validation!" << std::endl;
        std::size_t dot = file.rfind('.');
        if (dot == std::string::npos)
            throw BadFileFormatException();
        else throw BadFileFormatException(file.substr(file.rfind('.'), file.size() - file.rfind('.')));
    };
}

void Mesh_Loader(MeshLoader* loader_, const std::string& filepath_) 
{
    loader_->LoadMesh(filepath_);
    loader_->Print_Data();

    std::cout << "_______Result of Get_Elements_by_ID(2, 4, 8) function_______:\n";
    std::cout << loader_->Get_Elements_by_ID(2, 4, 8);

    std::cout << "\n______Result of Get_Elements_by_edge(2, 8) function_______:\n";
    std::cout << loader_->Get_Elements_by_rib(2, 8);

    std::cout << "\n_______Result of Get_Vertex_Nodes_by_Border_ID(6) function_______:\n";
    std::cout << loader_->Get_Vertex_Nodes_by_Border_ID(6);

	std::cout << "\n________Result of Get_Elements_by_ID_area(1) function_________:\n";
    std::cout << loader_->Get_Elements_by_ID_area(1);

    std::cout << "\n__________Result of Get_Borders_by_Border_ID(1) function________:\n";
    std::cout << loader_->Get_Borders_by_Border_ID(1);

	std::cout << "\n_________Result of Get_Nodes_Neighbors() function:_______\n";
    std::vector<std::set<int>> neighbors = loader_->Get_Nodes_Neighbors();
    for (int i = 0; i < neighbors.size(); ++i)
    {
        std::cout << "Neighbors for Node with ID = " << i + 1 << "\n" <<
            neighbors[i].size() << (neighbors[i].size() > 1 ? " Nodes:\n" : " Node:\n") << " ID" <<
            std::setw(10) << "X" << std::setw(11) << "Y" << std::setw(11) << "Z" << std::setw(14) << "Vertex\n";
        for (const auto& it : neighbors[i])
            std::cout << loader_->Get_Node_by_ID(it);
    }
    neighbors.clear();



	std::cout << "\n_______Result of Insert_New_Nodes() function:_______\n";
	loader_->Insert_New_Nodes();
	loader_->Print_Data();

    std::cout << "---------------Protection task---------------" << std::endl;
    Protection_task(*loader_);
}



int main(int argc, char* argv[]) 
{

	std::string filepath = "file.aneu";
	MeshLoader* loader = File_Format_Checker(filepath);
	Mesh_Loader(loader, filepath);

	delete loader;  
	return 0;
}
